@@ -1,6 +1,6 @@
 /* engine.c - GPGME engine support.
    Copyright (C) 2000 Werner Koch (dd9jn)
-   Copyright (C) 2001, 2002, 2003, 2004 g10 Code GmbH
+   Copyright (C) 2001, 2002, 2003, 2004, 2006 g10 Code GmbH
  
    This file is part of GPGME.
  
@@ -106,14 +106,33 @@ engine_get_req_version (gpgme_protocol_t proto)
 gpgme_error_t
 gpgme_engine_check_version (gpgme_protocol_t proto)
 {
+  gpgme_error_t err;
+  gpgme_engine_info_t info;
   int result;
-  char *engine_version = engine_get_version (proto, NULL);
 
-  result = _gpgme_compare_versions (engine_version,
-				    engine_get_req_version (proto));
-  if (engine_version)
-    free (engine_version);
+  LOCK (engine_info_lock);
+  info = engine_info;
+  if (!info)
+    {
+      /* Make sure it is initialized.  */
+      UNLOCK (engine_info_lock);
+      err = gpgme_get_engine_info (&info);
+      if (err)
+	return err;
 
+      LOCK (engine_info_lock);
+    }
+
+  while (info && info->protocol != proto)
+    info = info->next;
+
+  if (!info)
+    result = 0;
+  else
+    result = _gpgme_compare_versions (info->version,
+				      info->req_version);
+
+  UNLOCK (engine_info_lock);
   return result ? 0 : gpg_error (GPG_ERR_INV_ENGINE);
 }
 
@@ -335,7 +354,7 @@ _gpgme_set_engine_info (gpgme_engine_info_t info, gpgme_protocol_t proto,
   /* Install the new members.  */
   info->file_name = new_file_name;
   info->home_dir = new_home_dir;
-  info->version = engine_get_version (proto, file_name);
+  info->version = engine_get_version (proto, new_file_name);
 
   return 0;
 }
