@@ -119,6 +119,23 @@ DEFINE_STATIC_LOCK (writer_table_lock);
 
 
 
+static int
+get_desired_thread_priority (void)
+{
+  int value;
+
+  if (!_gpgme_get_conf_int ("IOThreadPriority", &value))
+    {
+      value = THREAD_PRIORITY_HIGHEST;
+      DEBUG1 ("** Using standard IOThreadPriority of %d\n", value);
+    }
+  else
+    DEBUG1 ("** Configured IOThreadPriority is %d\n", value);
+
+  return value;
+}
+
+
 static HANDLE
 set_synchronize (HANDLE h)
 {
@@ -266,7 +283,7 @@ create_reader (HANDLE fd)
       /* We set the priority of the thread higher because we know that
          it only runs for a short time.  This greatly helps to increase
          the performance of the I/O. */
-      SetThreadPriority (c->thread_hd, THREAD_PRIORITY_HIGHEST);
+      SetThreadPriority (c->thread_hd, get_desired_thread_priority ());
     }
 
     return c;
@@ -421,7 +438,13 @@ writer (void *arg)
     DEBUG2 ("writer thread %p for file %p started", c->thread_hd, c->file_hd );
     for (;;) {
         LOCK (c->mutex);
+        if ( c->stop_me ) {
+            UNLOCK (c->mutex);
+            break;
+        }
         if ( !c->nbytes ) { 
+            if (!SetEvent (c->is_empty))
+                DEBUG1 ("SetEvent failed: ec=%d", (int)GetLastError ());
             if (!ResetEvent (c->have_data) )
                 DEBUG1 ("ResetEvent failed: ec=%d", (int)GetLastError ());
             UNLOCK (c->mutex);
@@ -451,14 +474,6 @@ writer (void *arg)
       
         LOCK (c->mutex);
         c->nbytes -= nwritten;
-        if (c->stop_me) {
-            UNLOCK (c->mutex);
-            break;
-        }
-        if ( !c->nbytes ) {
-            if ( !SetEvent (c->is_empty) )
-                DEBUG1 ("SetEvent failed: ec=%d", (int)GetLastError ());
-        }
         UNLOCK (c->mutex);
     }
     /* indicate that we have an error  */
@@ -488,7 +503,7 @@ create_writer (HANDLE fd)
         return NULL;
 
     c->file_hd = fd;
-    c->have_data = CreateEvent (&sec_attr, FALSE, FALSE, NULL);
+    c->have_data = CreateEvent (&sec_attr, TRUE, FALSE, NULL);
     c->is_empty  = CreateEvent (&sec_attr, TRUE, TRUE, NULL);
     c->stopped = CreateEvent (&sec_attr, TRUE, FALSE, NULL);
     if (!c->have_data || !c->is_empty || !c->stopped ) {
@@ -524,7 +539,7 @@ create_writer (HANDLE fd)
       /* We set the priority of the thread higher because we know that
          it only runs for a short time.  This greatly helps to increase
          the performance of the I/O. */
-      SetThreadPriority (c->thread_hd, THREAD_PRIORITY_HIGHEST);
+      SetThreadPriority (c->thread_hd, get_desired_thread_priority ());
     }
 
     return c;
@@ -620,12 +635,14 @@ _gpgme_io_write ( int fd, const void *buffer, size_t count )
 
     LOCK (c->mutex);
     if ( c->nbytes ) { /* bytes are pending for send */
+        /* Reset the is_empty event.  Better safe than sorry.  */
+        if (!ResetEvent (c->is_empty))
+            DEBUG1 ("ResetEvent failed: ec=%d", (int)GetLastError ());
         UNLOCK (c->mutex);
         DEBUG2 ("fd %d: waiting for empty buffer in thread %p",
                 fd, c->thread_hd);
         WaitForSingleObject (c->is_empty, INFINITE);
         DEBUG2 ("fd %d: thread %p buffer is empty", fd, c->thread_hd);
-        assert (!c->nbytes);
         LOCK (c->mutex);
     }
     
@@ -634,11 +651,20 @@ _gpgme_io_write ( int fd, const void *buffer, size_t count )
         DEBUG1 ("fd %d: write error", fd );
         return -1;
     }
-      
+
+    /* If no error occured, the number of bytes in the buffer must be
+       zero.  */
+    assert (!c->nbytes);
+
     if (count > WRITEBUF_SIZE)
         count = WRITEBUF_SIZE;
     memcpy (c->buffer, buffer, count);
     c->nbytes = count;
+
+    /* We have to reset the is_empty event early, because it is also
+       used by the select() implementation to probe the channel.  */
+    if (!ResetEvent (c->is_empty))
+        DEBUG1 ("ResetEvent failed: ec=%d", (int)GetLastError ());
     if (!SetEvent (c->have_data))
         DEBUG1 ("SetEvent failed: ec=%d", (int)GetLastError ());
     UNLOCK (c->mutex);
@@ -843,7 +869,7 @@ _gpgme_io_spawn ( const char *path, char **argv,
     memset (&si, 0, sizeof si);
     si.cb = sizeof (si);
     si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
-    si.wShowWindow = debug_me? SW_SHOW : SW_MINIMIZE;
+    si.wShowWindow = debug_me? SW_SHOW : SW_HIDE;
     si.hStdInput = GetStdHandle (STD_INPUT_HANDLE);
     si.hStdOutput = GetStdHandle (STD_OUTPUT_HANDLE);
     si.hStdError = GetStdHandle (STD_ERROR_HANDLE);
@@ -942,60 +968,6 @@ _gpgme_io_spawn ( const char *path, char **argv,
 }
 
 
-
-
-int
-_gpgme_io_waitpid ( int pid, int hang, int *r_status, int *r_signal )
-{
-    HANDLE proc = fd_to_handle (pid);
-    int code, ret = 0;
-    DWORD exc;
-
-    *r_status = 0;
-    *r_signal = 0;
-    code = WaitForSingleObject ( proc, hang? INFINITE : 0 );
-    switch (code) {
-      case WAIT_FAILED:
-        DEBUG2 ("WFSO pid=%d failed: %d\n", (int)pid, (int)GetLastError () );
-        break;
-
-      case WAIT_OBJECT_0:
-        if (!GetExitCodeProcess (proc, &exc)) {
-            DEBUG2 ("** GECP pid=%d failed: ec=%d\n",
-                    (int)pid, (int)GetLastError () );
-            *r_status = 4; 
-        }
-        else {
-            DEBUG2 ("GECP pid=%d exit code=%d\n", (int)pid,  exc);
-            *r_status = exc;
-        }
-        ret = 1;
-        break;
-
-      case WAIT_TIMEOUT:
-        if (hang)
-            DEBUG1 ("WFSO pid=%d timed out\n", (int)pid);
-        break;
-
-      default:
-        DEBUG2 ("WFSO pid=%d returned %d\n", (int)pid, code );
-        break;
-    }
-    return ret;
-}
-
-int
-_gpgme_io_kill ( int pid, int hard )
-{
-    HANDLE proc = fd_to_handle (pid);
-
-    #warning I am not sure how to kill a process
-    /* fixme: figure out how this can be done */
-    return 0;
-}
-
-
-
 /*
  * Select on the list of fds.
  * Returns: -1 = error
@@ -1055,18 +1027,11 @@ _gpgme_io_select ( struct io_select_fd_s *fds, size_t nfds, int nonblock )
                         DEBUG0 ("Too many objects for WFMO!" );
                         return -1;
                     }
-                    LOCK (c->mutex);
-                    if ( !c->nbytes ) {
-                        waitidx[nwait]   = i;
-                        waitbuf[nwait++] = c->is_empty;
-                        DEBUG_ADD1 (dbg_help, "w%d ", fds[i].fd );
-                        any = 1;
-                    }
-                    else {
-                        DEBUG_ADD1 (dbg_help, "w%d(ignored) ", fds[i].fd );
-                    }
-                    UNLOCK (c->mutex);
+		    waitidx[nwait]   = i;
+		    waitbuf[nwait++] = c->is_empty;
                 }
+		DEBUG_ADD1 (dbg_help, "w%d ", fds[i].fd );
+		any = 1;
             }
         }
     }
@@ -1146,3 +1111,22 @@ _gpgme_io_subsystem_init (void)
 }
 
 
+/* Write the printable version of FD to the buffer BUF of length
+   BUFLEN.  The printable version is the representation on the command
+   line that the child process expects.  */
+int
+_gpgme_io_fd2str (char *buf, int buflen, int fd)
+{
+  return snprintf (buf, buflen, "%d", fd);
+}
+
+
+/* The following interface is only useful for GPGME Glib.  */
+
+/* Look up the giochannel for file descriptor FD.  */
+void *
+gpgme_get_giochannel (int fd)
+{
+  return NULL;
+}
+
